import os
import sys

# Attach $SUMO_HOME/tools to the path to import SUMO libraries
if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    import libsumo as traci
    # import traci    
    import traci.constants as tc
else:
    raise Exception("Please declare environment variable 'SUMO_HOME'")

class SUMOSim():
    def __init__(self, sumo_cfg_file):
        self.sumo_cfg_file = sumo_cfg_file
        self.sumo_params = ['sumo','-c',self.sumo_cfg_file]
        traci.start(self.sumo_params)

        self._is_ongoing = True
        # self._start_time = self.traci_handler.simulation.getTime()
        self._sumo_steps = 0
        self._manually_stopped = False
        self.traci_handler = traci

        self.traci_handler.simulation.subscribe([
            tc.VAR_COLLISIONS,
            tc.VAR_COLLIDING_VEHICLES_IDS,
            tc.VAR_COLLIDING_VEHICLES_NUMBER,
            # tc.LAST_STEP_VEHICLE_ID_LIST,
            # tc.LAST_STEP_VEHICLE_DATA,
        ])
        self.traci_handler.vehicletype.subscribe('veh', varIDs=[
            tc.VAR_ACCEL,
            tc.VAR_DECEL,
        ])
    
    def __del__(self):
        try:
            self.end_simulation()
        except KeyError:
            print('__del__: Simulation already closed.')

    def end_simulation(self):
        """ Forces the simulation to stop. """
        if self.is_ongoing_sim():
            self._manually_stopped = True
            self.traci_handler.close()
        else:
            print('end_simulation(): Simulation already closed')

    def is_ongoing_sim(self):
        """ Return True iff the SUMO simulation is still ongoing. """
        if self._manually_stopped:
            return False
        if self.traci_handler.simulation.getMinExpectedNumber() <= 0:
            # No entities left in the simulation.
            return False
        if self.traci_handler.simulation.getTime() > 3600:
            # the simulatio reach the predefined (from parameters) end
            return False
        return True

    def _stopping_condition(self, current_step_counter, until_end):
        """ Computes the stopping condition. """
        if self._manually_stopped:
            return True
        if self.traci_handler.simulation.getMinExpectedNumber() <= 0:
            # No entities left in the simulation.
            return True
        if self.traci_handler.simulation.getTime() > 3600:
            # the simulatio reach the predefined (from parameters) end
            return True
        if current_step_counter == 1 and not until_end:
            return True
        return False

    def _default_step_action(self, agents=None):
        """ Specific code to be executed in every simulation step """
        self.veh_subscriptions = self.traci_handler.vehicle.getAllSubscriptionResults()
        self.sim_subscriptions = self.traci_handler.simulation.getAllSubscriptionResults()
        
        # for veh, vals in self.veh_subscriptions.items():
        #     print('AGENT {}'.format(veh))
        #     print('VAR_SPEED : {}'.format(vals.get(tc.VAR_SPEED, "N/A")))
        #     print('VAR_ACCELERATION : {}'.format(vals.get(tc.VAR_ACCELERATION, "N/A")))
        #     print('VAR_LANEPOSITION : {}'.format(vals.get(tc.VAR_LANEPOSITION, "N/A")))
        #     print('VAR_LANE_INDEX : {}'.format(vals.get(tc.VAR_LANE_INDEX, "N/A")))
        #     print('VAR_LANE_ID : {}'.format(vals.get(tc.VAR_LANE_ID, "N/A")))
        #     print('VAR_MAXSPEED : {}'.format(vals.get(tc.VAR_MAXSPEED, "N/A")))
        #     print('VAR_PREV_SPEED : {}'.format(vals.get(tc.VAR_PREV_SPEED, "N/A")))
        #     print('VAR_ALLOWED_SPEED : {}'.format(vals.get(tc.VAR_ALLOWED_SPEED, "N/A")))
        #     print('LANECHANGE_LEFT : {}'.format(vals.get(tc.LANECHANGE_LEFT, "N/A")))
        #     print('LANECHANGE_RIGHT : {}'.format(vals.get(tc.LANECHANGE_RIGHT, "N/A")))
        #     print('LANE_ALLOWED : {}'.format(vals.get(tc.LANE_ALLOWED, "N/A")))
        #     print('LANE_CHANGES : {}'.format(vals.get(tc.LANE_CHANGES, "N/A")))
        #     print('LCA_WANTS_LANECHANGE : {}'.format(vals.get(tc.LCA_WANTS_LANECHANGE, "N/A")))
        #     print('LCA_WANTS_LANECHANGE_OR_STAY : {}'.format(vals.get(tc.LCA_WANTS_LANECHANGE_OR_STAY, "N/A")))
        #     print('VAR_COLLIDING_VEHICLES_IDS : {}'.format(vals.get(tc.VAR_COLLIDING_VEHICLES_IDS, "N/A")))
        #     print('VAR_COLLIDING_VEHICLES_NUMBER : {}'.format(vals.get(tc.VAR_COLLIDING_VEHICLES_NUMBER, "N/A")))
        #     print('VAR_COLLISIONS : {}'.format(vals.get(tc.VAR_COLLISIONS, "N/A")))
        #     print('VAR_ACCEL : {}'.format(vals.get(tc.VAR_ACCEL, "N/A")))
        #     print('VAR_DECEL : {}'.format(vals.get(tc.VAR_DECEL, "N/A")))
        
        # for sim, vals in self.sim_subscriptions.items():
            # print(sim)
            # print('LAST STEP VEHICLE ID LIST: {}'.format(vals.get(tc.LAST_STEP_VEHICLE_ID_LIST, "N/A")))
            # print('LAST STEP VEHICLE DATA: {}'.format(vals.get(tc.LAST_STEP_VEHICLE_DATA, "N/A")))
        # input()
        running = set()

        for agent in self.veh_subscriptions:
            running.add(agent)
        
        if len(running) == 0:
            print('All the agents left the simulation')
            self.end_simulation() # TRACI.CLOSE()
            # maybe use del here?


        return True

    def step(self, until_end=False, agents=set()):
        """
        Runs a "learning" step and returns if the simulation has finished.
        This function in meant to be called by the RLLIB Environment.

        Params:
            until_end: Bool. If True, run the sumo simulation until the end.
            agents: Set(String). It passes the agents to the _default_step_action function.

        Return:
            Bool. True iff the simulation is still ongoing.
        """
        ## Execute SUMO steps until the learning needs to happen
        current_step_counter = 0

        while not self._stopping_condition(current_step_counter, until_end):
            self.traci_handler.simulationStep()
            self._sumo_steps += 1
            current_step_counter += 1
            self._default_step_action(agents)

        # If the simulation has finished
        if self.is_ongoing_sim():
            return True
        return False

    def fast_forward(self, time):
        """
        Move the simulation forward (without doing anything else) until the given time.
        Param:
            time: Float, simulation time in seconds.
        """
        self.traci_handler.simulationStep(float(time))


class MyEnv():
    def __init__(self):
        self.sim = None
        self.episodes = 0
        self.reset()

    def subscribe_fn(self,vh):
        self.sim.traci_handler.vehicle.subscribe(vh, varIDs=[
            tc.VAR_SPEED,
            tc.VAR_ACCELERATION,
            tc.VAR_LANEPOSITION,
            tc.VAR_LANE_INDEX,
            tc.VAR_LANE_ID,
            tc.VAR_MAXSPEED,
        ])

    def unsubscribe_fn(self,vh):
        try:
            self.sim.traci_handler.vehicle.unsubscribe(vh)
        except:
            pass
    def reset(self):
        self.resetted = True
        self.episodes += 1
        self.steps = 0
        self.agents = {}

        # Reset the SUMO simulation
        if self.sim:
            del self.sim # TRACI.CLOSE()
            # self.sim.end_simulation()
            
        self.sim = SUMOSim('./scenario/sumo.cfg.xml') # # TRACI.START()/LOAD()
        
        while len(self.sim.traci_handler.vehicle.getIDList()) <= 0:
            self.sim.fast_forward(0)

        for vh in self.sim.traci_handler.vehicle.getIDList():
            self.agents[vh] = 1
            self.subscribe_fn(vh)

        # self.sim.step(self.agents.keys())
        self.sim._default_step_action(self.agents.keys())
        
        self.update_env_lane_zone_map()
        
        # obs = dict()
        # for id, agent in self.agents.items():
            # obs[id] = agent.observe(self.env_lane_zone_map, self.pheromones)
            # obs[id] = self.observation_space.sample()

        obs = dict()
        for id, agent in self.agents.items():
            obs[id] = [1,2,3]

        return obs

    def update_env_lane_zone_map(self):
        for agent in self.agents.keys():
            # lane_id = agent.veh.get_lane_id()
            lane_id = self.sim.veh_subscriptions[agent][tc.VAR_LANE_INDEX]
            # lane_id = self.sim.traci_handler.vehicle.getLaneIndex(agent)
            lane_pos = self.sim.veh_subscriptions[agent][tc.VAR_LANEPOSITION]

        # use lane id and lane pos to create map
            
        
    def step(self):
        self.resetted = False
        self.steps +=1 

        print('################## STEP {} ##########################'.format(self.steps))

        traci_vehicle_list = self.sim.traci_handler.vehicle.getIDList()
        print('Reference | TraCI getIDList(): {}'.format(traci_vehicle_list))
        
        subs_vehicle_list = self.sim.veh_subscriptions.keys()
        print('Vehicle subscriptions: {}'.format(list(subs_vehicle_list)))

        agents_vehicle_list = self.agents.keys()
        print('Agents: {}'.format(list(self.agents.keys())))

        # MAKE VEHICLES INSERTED IN SIMLATION INTO AGENTS AND SUBSCRIBE
        for vh in traci_vehicle_list:
            if vh not in agents_vehicle_list:
                self.agents[vh] = 1
                self.subscribe_fn(vh)

        print('############ New vehicles inserted are agents now and subscribed ############')
        subs_vehicle_list = self.sim.veh_subscriptions.keys()
        print('Vehicle subscriptions: {}'.format(list(subs_vehicle_list)))

        agents_vehicle_list = self.agents.keys()
        print('Agents: {}'.format(list(self.agents.keys())))

        print('-----------------------------------------------------------------------------')

        agents_to_remove = []
        for vh in agents_vehicle_list:
            if vh not in traci_vehicle_list:
                agents_to_remove.append(vh)

        for vh in agents_to_remove:
            self.unsubscribe_fn(vh)
            self.agents.pop(vh)

        print('###### Vehicles leaving are removed from -agent- status and unsubscribed ######')

        subs_vehicle_list = self.sim.veh_subscriptions.keys()
        print('Vehicle subscriptions: {}'.format(list(subs_vehicle_list)))

        agents_vehicle_list = self.agents.keys()
        print('Agents: {}'.format(list(self.agents.keys())))

        print('-------------------------------------------------------------------------------')

        input()
        self.update_env_lane_zone_map()



        # for agent_id, agent in self.agents.items():
            # agent.apply_action(action_dict[agent_id]) ### PROBLEM USING CHANGELANERELATIVE HERE SOMEHOW. SYNC PROBLEM? NETWORK NOT YET CONSTRUCTED

        # Gather observations for each agent - Not relevant right now
        observations = dict()
        rewards = dict()

        for agent_id, agent in self.agents.items():
            observations[agent_id] = [1,2,3]
            rewards[agent_id] = -len(self.sim.traci_handler.simulation.getCollidingVehiclesIDList())

        ongoing_simulation = self.sim.step(until_end=False, agents=self.agents.keys())
        ## end of the episode
        if not ongoing_simulation:
            print('Reached the end of the SUMO simulation.')
            self.reset()

        return observations, rewards



if __name__ == '__main__':
    env = MyEnv()
    for i in range(100): # similar to episodes
        sim_time = 0
        while sim_time <= 150: # similar to steps
            env.step()
            sim_time += 1
